/*
 * tio - a simple serial terminal I/O tool
 *
 * Copyright (c) 2022  Martin Lund
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <errno.h>
#include "error.h"
#include "print.h"
#include "misc.h"
#include "tty.h"

static void signal_handler(int signum)
{
    switch (signum)
    {
        case SIGHUP:
            tio_printf("Received SIGHUP signal!");
            break;
        case SIGINT:
            tio_printf("Received SIGINT signal!");
            break;
    }
    exit(EXIT_FAILURE);
}

void signal_handlers_install(void)
{
    signal(SIGHUP, signal_handler);
    signal(SIGINT, signal_handler);
    signal(SIGPIPE, SIG_IGN);
}
